#ifndef _SQLIMPORT_CPP
#define _SQLIMPORT_CPP
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <Windows.H>
#include <WindowsX.H>
#include <ShellAPI.H>
#include <Stdio.H>
#include <Stdlib.H>
#include <SQL.H>
#include <SQLExt.H>

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "../Resources/Resource.H"
#include "../../SharedClasses/CRC32/CRC.H"
#include "../../SharedClasses/SQLClass/cSQL.H"
#include "../../SharedClasses/SQLClass/cRecordSet.H"
#include "../SockServer/SockServer.H"

#include "NSWFL.H"
#include "Init.H"
#include "Entry.H"
#include "Routines.H"
#include "Command.H"
#include "SQLImport.H"

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

SQLINTEGER CharTrim(char *Data, int Length)
{
    SQLINTEGER Pos = Length;

    if(Length == 0)
        return Length;

    if(Data[Length-1] != ' ')
        return Length;

    Pos--;

    while(Pos != 0 && Data[Pos] == ' ')
        Pos--;

    if(Data[Pos] != ' ')
        Pos++;

    Data[Pos] = '\0';

    return Pos;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void FreeSingleRowDataColumbs(struct _SQLImportMembers *SIM, int ColumnCount)
{
    int ColumbPos = 0;
    while(ColumbPos < ColumnCount)
    {
        free(SIM->SingleRowData[ColumbPos]);
        SIM->SingleRowData[ColumbPos] = NULL;
        ColumbPos++;
    }
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void FreeColumbNames(struct _SQLImportMembers *SIM, int ColumnCount)
{
    int ColumbPos = 0;
    while(ColumbPos < ColumnCount)
    {
        free(SIM->ColumbNames[ColumbPos]);
        SIM->ColumbNames[ColumbPos] = NULL;
        ColumbPos++;
    }
    free(SIM->ColumbNames);
    SIM->ColumbNames = NULL;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void FreeSingleRowData(struct _SQLImportMembers *SIM)
{
    free(SIM->SingleRowData);
    SIM->SingleRowData = NULL;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

char **GetPrimaryKeys(CSQL *pcSQL, char *sTableName, int *iNumOfKeys)
{
	int iPKCount = 0;
	int iPKColLen = 0;

	char sSQL[1024];
	char sPKCol[256];
	char *sBaseSQL = NULL;
	char **sOutKeys = NULL;

    CRecordSet rsTemp;

	sBaseSQL = "SELECT sysobjects.parent_obj AS tableid, 'PK' AS constraintType,"
		" sysobjects.name AS constraintName, syscolumns.name AS keyColumn, sysindexkeys.KeyNo"
		" FROM sysobjects"
		" INNER JOIN sysindexes ON sysobjects.name = sysindexes.name"
		" INNER JOIN sysindexkeys ON sysindexes.indid = sysindexkeys.indid AND sysindexes.id = sysindexkeys.id"
		" INNER JOIN syscolumns ON sysindexkeys.colid = syscolumns.colid AND sysobjects.parent_obj = syscolumns.id"
		" WHERE sysobjects.xtype = 'PK' and sysobjects.parent_obj = object_id('%s')"
		" Order by sysindexkeys.KeyNo";

	sprintf(sSQL, sBaseSQL, sTableName);

	if(!pcSQL->Execute(sSQL, &rsTemp))
	{
		return NULL;
	}

	if(rsTemp.RowCount > 0)
	{
		sOutKeys = (char **) calloc(sizeof(char *), rsTemp.RowCount + 1);

		while(rsTemp.Fetch())
		{
			rsTemp.sColumnEx(4, sPKCol, sizeof(sPKCol), &iPKColLen);

			if(iPKColLen > 0)
			{
				sOutKeys[iPKCount] = (char *) calloc(sizeof(char), iPKColLen + 1);
				strcpy(sOutKeys[iPKCount], sPKCol);
				iPKCount++;
			}
		}
		sOutKeys[iPKCount] = NULL;
	}
	else{
		sOutKeys = NULL;
	}

	rsTemp.Close();
	
	*iNumOfKeys = iPKCount;

	return sOutKeys;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void FreePrimaryKeys(char **sOutKeys, int iKeys)
{
	int iPos = 0;

	if(iKeys > 0)
	{
		while(iPos < iKeys)
		{
			free(sOutKeys[iPos]);
			iPos++;
		}
	}
	else{
		while(sOutKeys[iPos])
		{
			free(sOutKeys[iPos]);
			iPos++;
		}
	}

	free(sOutKeys);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool ImportSQLResults(SocketServer *pSockSrvr, int iClient, char *InFileName)
{
    char StatusText[255];

    int outDataLength = 0;
	int outRowCount = 0;

    FILE *SourceHandle = NULL;

    _SQLImportMembers SIM;

    CRecordSet rsTemp;

	WriteLog(pSockSrvr->icClientID[iClient], "Importing client data.");

    if((SourceHandle = fopen(InFileName, "rb")) == NULL)
    {
        // FIXME: Need to dump the file name to log
		WriteLog(pSockSrvr->icClientID[iClient], "Error in ImportSQLResultsFailed:: TargetFile open error.");
        return false;
    }

    int TotalLengthOfAllColumbNames = 0; // For better memory management.
    int TotalLengthOfAllRowData = 0; // For better memory management.
    int ImportTableNameLen = 0;
    int ColumnCount = 0;
    int RowCount = 0;
    char ImportTable[256];

    fread(&ImportTableNameLen, sizeof(ImportTableNameLen), 1, SourceHandle); // Read import table name len
    fread(ImportTable, sizeof(char), ImportTableNameLen, SourceHandle); // Read import table name
    fread(&ColumnCount, sizeof(ColumnCount), 1, SourceHandle); // Read number of columbs
    fread(&RowCount, sizeof(RowCount), 1, SourceHandle);       // Read number of rows

    ImportTable[ImportTableNameLen] = '\0';

    char **sUniuqeKeyNames = NULL;   // Contains the names of the uniuqe columbs.
    int *iUniuqeKeyCols = NULL;      // Contains the columb position of the uniuqe columbs.
    int iUniuqeKeysNeeded = 0;       // The ammount of uniuqe keys we have to find.
    int iUniuqeKeysFound = 0;        // The ammount of uniuqe keys we have found.

	sUniuqeKeyNames = GetPrimaryKeys(&CCI[iClient].cCustSQL, ImportTable, &iUniuqeKeysNeeded);

	if(&iUniuqeKeysNeeded > 0)
	{
		iUniuqeKeyCols = (int *) calloc(sizeof(int), iUniuqeKeysNeeded);
	}
	else{
		//No primary keys, cannot import this data.
		WriteLog(pSockSrvr->icClientID[iClient], "Error in ImportSQLResultsFailed:: Could not obtain a list of primary keys.");

        if(SourceHandle) fclose(SourceHandle);
        return false;
	}

    SIM.ColumbNames = (char **) calloc(sizeof(char *), ColumnCount + 1);

    int ColNameLen = 0;

    int ColumbPos = 0;
    while(ColumbPos < ColumnCount)
    {
        fread(&ColNameLen, sizeof(ColNameLen), 1, SourceHandle);
        SIM.ColumbNames[ColumbPos] = (char *) calloc(sizeof(char), ColNameLen + 1);
        fread(SIM.ColumbNames[ColumbPos], sizeof(char), ColNameLen, SourceHandle);
        SIM.ColumbNames[ColumbPos][ColNameLen] = '\0';
        TotalLengthOfAllColumbNames = (TotalLengthOfAllColumbNames + ColNameLen);

        // Is this columb one of the uniuqe keys
        if(iUniuqeKeysFound < iUniuqeKeysNeeded)
        {
            int UniuqeKeyCount = 0;
            while(UniuqeKeyCount < iUniuqeKeysNeeded)
            {
                if(strcmpi(sUniuqeKeyNames[UniuqeKeyCount], SIM.ColumbNames[ColumbPos]) == 0)
                {
                    iUniuqeKeyCols[UniuqeKeyCount] = ColumbPos;
                    iUniuqeKeysFound++;
                }
                UniuqeKeyCount++;
            }
        }

        ColumbPos++;
    }

	int UniuqeKeyCount = 0;
    while(iUniuqeKeysNeeded < iUniuqeKeysFound)
    {
        sprintf(StatusText, "Error in ImportSQLResultsFailed:: Required uniuqe keys wer not found for: '%s'.", ImportTable);
		WriteLog(pSockSrvr->icClientID[iClient], StatusText);

        FreeColumbNames(&SIM, ColumnCount);

        if(SourceHandle) fclose(SourceHandle);
        return false;
    }

    if( pSockSrvr->bcConnected[iClient] == false )
    {
		WriteLog(pSockSrvr->icClientID[iClient], "Disconnected before data import.");

        if(SourceHandle) fclose(SourceHandle);
        return false;
    }

	WriteLog(pSockSrvr->icClientID[iClient], "Starting row by row data import.");

    SIM.SingleRowData = (char **) calloc(sizeof(char *), ColumnCount + 1);

    int RowDataLen = 0;

    int RowPos = 0;
    while(RowPos < RowCount)
    {
        int BaseMemoryRequirement = 0;
        TotalLengthOfAllRowData = 0;

        // Get row data
        ColumbPos = 0;
        while(ColumbPos < ColumnCount)
        {
            fread(&RowDataLen, sizeof(RowDataLen), 1, SourceHandle);
            SIM.SingleRowData[ColumbPos] = (char *) calloc(sizeof(char), RowDataLen + 1);
            fread(SIM.SingleRowData[ColumbPos], sizeof(char), RowDataLen, SourceHandle);
            SIM.SingleRowData[ColumbPos][RowDataLen] = '\0';
            TotalLengthOfAllRowData = (TotalLengthOfAllRowData + RowDataLen);
            ColumbPos++;
        }

        BaseMemoryRequirement = ((TotalLengthOfAllColumbNames + TotalLengthOfAllRowData) + (ColumnCount * 5)) + 1024;

        char WhereStatement[1024];
        char CheckExistanceStatement[1024];

        // Assemble the where statement from the row data, the uniuqe key names and positions
        if(iUniuqeKeysFound > 0)
        {
            // Assemble the where statement
            sprintf(WhereStatement, "WHERE %s = %s", sUniuqeKeyNames[0], SIM.SingleRowData[iUniuqeKeyCols[0]]);
        	int UniuqeKeyCount = 1;
            while(UniuqeKeyCount < iUniuqeKeysFound)
            {
                strcat(WhereStatement, " AND ");
                strcat(WhereStatement, sUniuqeKeyNames[UniuqeKeyCount]);
                strcat(WhereStatement, " = ");
                strcat(WhereStatement, SIM.SingleRowData[iUniuqeKeyCols[UniuqeKeyCount]]);
                UniuqeKeyCount++;
            }

            // Assemble the check existance statement
            sprintf(CheckExistanceStatement, "SELECT %s", sUniuqeKeyNames[0]);
            strcat(CheckExistanceStatement, " FROM ");
            strcat(CheckExistanceStatement, ImportTable);
            strcat(CheckExistanceStatement, " ");
            strcat(CheckExistanceStatement, WhereStatement);
        }

		if(!CCI[iClient].cCustSQL.Execute(CheckExistanceStatement, &rsTemp))
        {
            // FIXME: Need to dump the statement that failed to a seperate log file
			WriteLog(pSockSrvr->icClientID[iClient], "Error in ImportSQLResultsFailed:: Execute failed on CheckExistanceStatement.");

            FreeSingleRowDataColumbs(&SIM, ColumnCount);
            FreeSingleRowData(&SIM);
            FreeColumbNames(&SIM, ColumnCount);

            if(SourceHandle) fclose(SourceHandle);
            return false;
        }

        int iCountOfRowsThatAlreadyExist = rsTemp.RowCount;

        rsTemp.Close();

        char *ImportStatement = NULL;
        bool FreeImportStatement = false;

        //----------------------------------------------------------------------------------------------
        if(iCountOfRowsThatAlreadyExist == 0) // Do insert
        {
            ImportStatement = (char *) calloc( sizeof(char), BaseMemoryRequirement + 1);
            FreeImportStatement = true;

            sprintf(ImportStatement, "INSERT INTO %s (", ImportTable);

            ColumbPos = 0;
            while(ColumbPos < ColumnCount)
            {
                strcat(ImportStatement, SIM.ColumbNames[ColumbPos]);
                if(ColumbPos < ColumnCount - 1)
                {
                    strcat(ImportStatement, ",");
                }
                ColumbPos++;
            }
            strcat(ImportStatement, ") VALUES(");

            ColumbPos = 0;
            while(ColumbPos < ColumnCount)
            {
                strcat(ImportStatement, SIM.SingleRowData[ColumbPos]);
                if(ColumbPos < ColumnCount - 1)
                {
                    strcat(ImportStatement, ",");
                }
                ColumbPos++;
            }
            strcat(ImportStatement, ") ");
        }
        //----------------------------------------------------------------------------------------------
        else { // Do update
            BaseMemoryRequirement = (BaseMemoryRequirement + strlen(WhereStatement));
            ImportStatement = (char *) calloc( sizeof(char), BaseMemoryRequirement + 1);
            FreeImportStatement = true;

            sprintf(ImportStatement, "UPDATE %s SET ", ImportTable);

            ColumbPos = 0;
            while(ColumbPos < ColumnCount)
            {
                strcat(ImportStatement, SIM.ColumbNames[ColumbPos]);
                strcat(ImportStatement, "=");
                strcat(ImportStatement, SIM.SingleRowData[ColumbPos]);
                if(ColumbPos < ColumnCount - 1)
                {
                    strcat(ImportStatement, ",");
                }
                ColumbPos++;
            }

            strcat(ImportStatement, " ");
            strcat(ImportStatement, WhereStatement);
        }

        //----------------------------------------------------------------------------------------------

        if(ImportStatement != NULL)
        {
			if(!CCI[iClient].cCustSQL.Execute(ImportStatement, &rsTemp))
            {
                // FIXME: Need to dump the statement that failed to a seperate log file. (ImportStatement)
				WriteLog(pSockSrvr->icClientID[iClient], "Error in ImportSQLResultsFailed:: Execute failed on ImportStatement.");

                FreeSingleRowDataColumbs(&SIM, ColumnCount);
                FreeSingleRowData(&SIM);
                FreeColumbNames(&SIM, ColumnCount);

                free(ImportStatement);

                if(SourceHandle) fclose(SourceHandle);
                return false;
            }

            rsTemp.Close();
        }

        if(FreeImportStatement)
		{
            free(ImportStatement);
		}

        ImportStatement = NULL;

        FreeSingleRowDataColumbs(&SIM, ColumnCount);
        //FreeSingleRowData(&SIM); // This memory needs to stay intact untill we are done with the import.

        ColumbPos++;

        RowPos++;

        if( pSockSrvr->bcConnected[iClient] == false )
        {
		    sprintf(StatusText, "Disconnected durring data import, after %d successful rows.", RowPos);
			WriteLog(pSockSrvr->icClientID[iClient], StatusText);

            FreeSingleRowData(&SIM);
            FreeColumbNames(&SIM, ColumnCount);

            if(SourceHandle) fclose(SourceHandle);
            return false;
        }
    }

    sprintf(StatusText, "Successfully imported %d client data rows.", RowPos);
	WriteLog(pSockSrvr->icClientID[iClient], StatusText);

	if(iUniuqeKeysNeeded > 0)
	{
		FreePrimaryKeys(sUniuqeKeyNames, iUniuqeKeysNeeded);
		free(iUniuqeKeyCols);
	}

	FreeSingleRowData(&SIM);
    FreeColumbNames(&SIM, ColumnCount); 

    if(SourceHandle) fclose(SourceHandle);

    return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#endif
